"""App and View testing."""
import pytest

from bowtie import App
from bowtie.control import Button


def test_subscribe_error():
    """Subscribe with incorrect argument order."""
    app = App()
    button = Button()
    with pytest.raises(TypeError):
        app.subscribe(3, button.on_click)



"""Bowtie abstract component classes.

All visual and control components inherit these.
"""

from typing import Any, Callable, Optional, ClassVar, Tuple  # pylint: disable=unused-import
from abc import ABCMeta, abstractmethod
import string
from functools import wraps
import json
from datetime import datetime, date, time

import msgpack
import flask
from flask_socketio import emit
import eventlet
from eventlet.queue import LightQueue

from bowtie.exceptions import SerializationError
from bowtie._typing import JSON


COMPONENT_REGISTRY = {}
SEPARATOR = '#'


class Event:
    """Data structure to hold information for events."""

    def __init__(self, name: str, uuid: int, getter: Optional[str] = None) -> None:
        """Create an event.

        Parameters
        ----------
        name : str
        uuid : int
        getter : str, optional

        """
        self.name = name
        self.uuid = uuid
        self.getter = getter

    @property
    def signal(self) -> str:
        """Name of socket.io message."""
        return '{}{}{}'.format(self.uuid, SEPARATOR, self.name)

    @property
    def _key(self) -> Tuple[str, int, Optional[str]]:
        return self.name, self.uuid, self.getter

    def __repr__(self) -> str:
        """Create an Event."""
        return "Event('{}', {}, '{}')".format(self.name, self.uuid, self.getter)

    def __eq__(self, other) -> bool:
        """Compare Events for equality."""
        # pylint: disable=protected-access
        return isinstance(other, type(self)) and self._key == other._key

    def __hash__(self) -> int:
        """Compute hash for Event."""
        return hash(self._key)


def jsbool(x: bool) -> str:
    """Convert Python bool to Javascript bool."""
    return repr(x).lower()


def json_conversion(obj: Any) -> JSON:
    """Encode additional objects to JSON."""
    try:
        # numpy isn't an explicit dependency of bowtie
        # so we can't assume it's available
        import numpy as np
        if isinstance(obj, (np.ndarray, np.generic)):
            return obj.tolist()
    except ImportError:
        pass

    try:
        # pandas isn't an explicit dependency of bowtie
        # so we can't assume it's available
        import pandas as pd
        if isinstance(obj, pd.DatetimeIndex):
            return [x.isoformat() for x in obj.to_pydatetime()]
        if isinstance(obj, pd.Index):
            return obj.tolist()
        if isinstance(obj, pd.Series):
            try:
                return [x.isoformat() for x in obj.dt.to_pydatetime()]
            except AttributeError:
                return obj.tolist()
    except ImportError:
        pass

    if isinstance(obj, (datetime, time, date)):
        return obj.isoformat()
    raise TypeError('Not sure how to serialize {} of type {}'.format(obj, type(obj)))


def jdumps(data: Any) -> str:
    """Encode Python object to JSON with additional encoders."""
    return json.dumps(data, default=json_conversion)


def encoders(obj: Any) -> JSON:  # pylint: disable=too-many-return-statements
    """Convert Python object to msgpack encodable ones."""
    try:
        # numpy isn't an explicit dependency of bowtie
        # so we can't assume it's available
        import numpy as np
        if isinstance(obj, (np.ndarray, np.generic)):
            # https://docs.scipy.org/doc/numpy/reference/arrays.scalars.html
            return obj.tolist()
    except ImportError:
        pass

    try:
        # pandas isn't an explicit dependency of bowtie
        # so we can't assume it's available
        import pandas as pd
        if isinstance(obj, pd.DatetimeIndex):
            return [x.isoformat() for x in obj.to_pydatetime()]
        if isinstance(obj, pd.Index):
            return obj.tolist()
        if isinstance(obj, pd.Series):
            try:
                return [x.isoformat() for x in obj.dt.to_pydatetime()]
            except AttributeError:
                return obj.tolist()
    except ImportError:
        pass

    if isinstance(obj, (datetime, time, date)):
        return obj.isoformat()

    return obj


def pack(x: Any) -> bytes:
    """Encode ``x`` into msgpack with additional encoders."""
    try:
        return msgpack.packb(x, default=encoders)
    except TypeError as exc:
        message = ('Serialization error, check the data passed to a do_ command. '
                   'Cannot serialize this object:\n') + str(exc)[16:]
        raise SerializationError(message)


def unpack(x: bytes) -> JSON:
    """Decode ``x`` from msgpack into Python object."""
    return msgpack.unpackb(x, encoding='utf8')


def make_event(event: Callable) -> Callable:
    """Create an event from a method signature."""
    @property  # type: ignore
    @wraps(event)
    def actualevent(self):  # pylint: disable=missing-docstring
        name = event.__name__[3:]
        try:
            # the getter post processing function
            # is preserved with an underscore
            getter = event(self).__name__
        except AttributeError:
            getter = None
        return Event(name, self._uuid, getter)  # pylint: disable=protected-access

    return actualevent


def is_event(attribute: str) -> bool:
    """Test if a method is an event."""
    return attribute.startswith('on_')


def make_command(command: Callable) -> Callable:
    """Create an command from a method signature."""
    @wraps(command)
    def actualcommand(self, *args, **kwds):  # pylint: disable=missing-docstring
        data = command(self, *args, **kwds)
        name = command.__name__[3:]
        signal = '{uuid}{sep}{event}'.format(
            uuid=self._uuid,  # pylint: disable=protected-access
            sep=SEPARATOR,
            event=name
        )
        if flask.has_request_context():
            emit(signal, {'data': pack(data)})
        else:
            sio = flask.current_app.extensions['socketio']
            sio.emit(signal, {'data': pack(data)})
        eventlet.sleep()

    return actualcommand


def is_command(attribute: str) -> bool:
    """Test if a method is an command."""
    return attribute.startswith('do_')


def make_getter(getter: Callable) -> Callable:
    """Create an command from a method signature."""
    def get(self, timeout=10):  # pylint: disable=missing-docstring
        name = getter.__name__
        signal = '{uuid}{sep}{event}'.format(
            uuid=self._uuid,  # pylint: disable=protected-access
            sep=SEPARATOR,
            event=name
        )
        event = LightQueue(1)
        if flask.has_request_context():
            emit(signal, callback=lambda x: event.put(unpack(x)))
        else:
            sio = flask.current_app.extensions['socketio']
            sio.emit(signal, callback=lambda x: event.put(unpack(x)))
        data = event.get(timeout=timeout)
        return getter(self, data)

    # don't want to copy the signature in this case
    get.__doc__ = getter.__doc__

    return get


def is_getter(attribute: str) -> bool:
    """Test if a method is a getter.

    It can be `get` or `get_*`.
    """
    return attribute.startswith('get')


class _Maker(ABCMeta):
    def __new__(mcs, name, bases, namespace):  # pylint: disable=arguments-differ
        for k in list(namespace.keys()):
            if is_event(k):
                namespace[k] = make_event(namespace[k])
            if is_command(k):
                namespace[k] = make_command(namespace[k])
            if is_getter(k):
                # preserve the post-processor with an underscore
                namespace['_' + k] = namespace[k]
                namespace[k] = make_getter(namespace[k])
        return super().__new__(mcs, name, bases, namespace)


class FormatDict(dict):
    """Dict to replace missing keys."""

    def __missing__(self, key: str) -> str:
        """Replace missing key with '{key}'."""
        return '{' + key + '}'


class Component(metaclass=_Maker):  # pylint: disable=too-few-public-methods
    """Abstract class for all components.

    All visual and control classes subclass this so their events
    and commands get transformed by the metaclass.
    """

    _NEXT_UUID = 0  # type: ClassVar[int]

    @property
    @abstractmethod
    def _TEMPLATE(self): pass  # pylint: disable=invalid-name,multiple-statements

    @property
    @abstractmethod
    def _COMPONENT(self): pass  # pylint: disable=invalid-name,multiple-statements

    @property
    @abstractmethod
    def _PACKAGE(self): pass  # pylint: disable=invalid-name,multiple-statements

    @property
    @abstractmethod
    def _ATTRS(self): pass  # pylint: disable=invalid-name,multiple-statements

    @property
    @abstractmethod
    def _instantiate(self): pass  # pylint: disable=invalid-name,multiple-statements

    @classmethod
    def _next_uuid(cls) -> int:
        cls._NEXT_UUID += 1
        return cls._NEXT_UUID

    def __init__(self) -> None:
        """Give the component a unique ID."""
        # wanted to put "self" instead of "Component"
        # was surprised that didn't work
        self._uuid = Component._next_uuid()
        super().__init__()
        self._tagbase = " socket={{socket}} uuid={{'{uuid}'}} />".format(uuid=self._uuid)
        self._tag = '<' + self._COMPONENT
        if self._ATTRS:
            self._tag += ' ' + self._ATTRS
        self._comp = None
        COMPONENT_REGISTRY[self._uuid] = self

    @staticmethod
    def _insert(wrap: str, tag: Optional[str]) -> str:
        """Insert the component tag into the wrapper html.

        This ignores other tags already created like ``{socket}``.

        https://stackoverflow.com/a/11284026/744520
        """
        if tag is None:
            raise ValueError('tag cannot be None')
        formatter = string.Formatter()
        mapping = FormatDict(component=tag)
        return formatter.vformat(wrap, (), mapping)

    def __eq__(self, other) -> bool:
        """Compare Events for equality."""
        # pylint: disable=protected-access
        return isinstance(other, type(self)) and self._uuid == other._uuid

    def __hash__(self) -> int:
        """Compute hash for Event."""
        return hash(self._uuid)



"""
Decorates a function for Bowtie.

Reference
---------
https://gist.github.com/carlsmith/800cbe3e11f630ac8aa0
"""

from typing import Callable
import os
import sys
import inspect
from subprocess import call

import click

from bowtie._app import _DIRECTORY, _WEBPACK, App


class WrongNumberOfArguments(TypeError):
    """The "build" function accepts an incorrect number of arguments."""

    pass


def numargs(func: Callable) -> int:
    """Get number of arguments."""
    return len(inspect.signature(func).parameters)


def _build(app):
    if app is None:
        print(('No `App` instance was returned. '
               'In the function decorated with @command, '
               'return the `App` instance so it can be built.'))
    else:
        if not isinstance(app, App):
            raise Exception(('Returned value {} is of type {}, '
                             'it needs to be a bowtie.App instance.'.format(app, type(app))))
        # pylint:disable=protected-access
        app._build()


def command(func):
    """Command line interface decorator.

    Decorate a function for building a Bowtie
    application and turn it into a command line interface.
    """
    @click.group(options_metavar='[--help]')
    def cmd():
        """Bowtie CLI to help build and run your app."""
        pass

    # pylint: disable=unused-variable
    @cmd.command(add_help_option=False)
    def build():
        """Write the app, downloads the packages, and bundles it with Webpack."""
        nargs = numargs(func)
        if nargs == 0:
            app = func()
        else:
            raise WrongNumberOfArguments(
                'Decorated function "{}" should have no arguments, it has {}.'
                .format(func.__name__, nargs)
            )
        _build(app)

    @cmd.command(context_settings=dict(ignore_unknown_options=True),
                 add_help_option=False)
    @click.argument('extra', nargs=-1, type=click.UNPROCESSED)
    def run(extra):
        """Build the app and serve it."""
        nargs = numargs(func)
        if nargs == 0:
            app = func()
        else:
            raise WrongNumberOfArguments(
                'Decorated function "{}" should have no arguments, it has {}.'
                .format(func.__name__, nargs)
            )
        # pylint:disable=protected-access
        _build(app)
        filepath = './{}/src/server.py'.format(_DIRECTORY)
        line = (filepath,) + extra
        call(line)

    @cmd.command(context_settings=dict(ignore_unknown_options=True),
                 add_help_option=False)
    @click.argument('extra', nargs=-1, type=click.UNPROCESSED)
    def serve(extra):
        """Serve the Bowtie app."""
        filepath = './{}/src/server.py'.format(_DIRECTORY)
        if os.path.isfile(filepath):
            line = (filepath,) + extra
            call(line)
        else:
            print("Cannot find '{}'. Did you build the app?".format(filepath))

    @cmd.command(context_settings=dict(ignore_unknown_options=True),
                 add_help_option=False)
    @click.argument('extra', nargs=-1, type=click.UNPROCESSED)
    def dev(extra):
        """Recompile the app for development."""
        line = (_WEBPACK, '--config', 'webpack.dev.js') + extra
        call(line, cwd=_DIRECTORY)

    @cmd.command(context_settings=dict(ignore_unknown_options=True),
                 add_help_option=False)
    @click.argument('extra', nargs=-1, type=click.UNPROCESSED)
    def prod(extra):
        """Recompile the app for production."""
        line = (_WEBPACK, '--config', 'webpack.prod.js', '--progress') + extra
        call(line, cwd=_DIRECTORY)

    locale = inspect.stack()[1][0].f_locals
    module = locale.get("__name__")

    if module == "__main__":
        try:
            arg = sys.argv[1:]
        except IndexError:
            arg = ('--help',)
        # pylint: disable=too-many-function-args
        sys.exit(cmd(arg))

    return cmd



#!/usr/bin/env python
"""Test all components for instatiation issues."""

import os
from os import environ as env
import subprocess
from inspect import isclass
import time

import pytest
from bowtie import App
from bowtie import control, visual, html
from bowtie._component import COMPONENT_REGISTRY
from bowtie.tests.utils import reset_uuid


def create_components():
    """Create components for this test."""
    reset_uuid()
    # pylint: disable=protected-access
    controllers = []
    for compstr in dir(control):
        comp = getattr(control, compstr)
        if (compstr[0] != '_' and isclass(comp) and issubclass(comp, control._Controller)
                and compstr != 'Upload'):
            controllers.append(comp())

    for controller in controllers:
        assert COMPONENT_REGISTRY[controller._uuid] == controller

    visuals = []
    for compstr in dir(visual):
        comp = getattr(visual, compstr)
        if compstr[0] != '_' and isclass(comp) and issubclass(comp, visual._Visual):
            visuals.append(comp())

    for vis in visuals:
        assert COMPONENT_REGISTRY[vis._uuid] == vis

    htmls = []
    for compstr in dir(html):
        comp = getattr(html, compstr)
        if compstr[0] != '_' and isclass(comp) and issubclass(comp, html._HTML):
            htmls.append(comp())

    for htm in htmls:
        assert COMPONENT_REGISTRY[htm._uuid] == htm

    return controllers, visuals, htmls


create_components()


@pytest.fixture
def components(build_path, monkeypatch):
    """App with all components."""
    monkeypatch.setattr(App, '_sourcefile', lambda self: 'bowtie.tests.test_components')

    controllers, visuals, htmls = create_components()

    app = App(rows=len(visuals))
    for controller in controllers:
        # pylint: disable=protected-access
        assert COMPONENT_REGISTRY[controller._uuid] == controller
        app.add_sidebar(controller)

    for vis in visuals:
        # pylint: disable=protected-access
        assert COMPONENT_REGISTRY[vis._uuid] == vis
        app.add(vis)

    for htm in htmls:
        # pylint: disable=protected-access
        assert COMPONENT_REGISTRY[htm._uuid] == htm
        app.add_sidebar(htm)

    assert len(COMPONENT_REGISTRY) == len(controllers) + 2 * len(visuals) + len(htmls)

    # pylint: disable=protected-access
    app._build()

    # run second time to make sure nothing weird happens with subsequent builds
    app._build()

    env['PYTHONPATH'] = '{}:{}'.format(os.getcwd(), os.environ.get('PYTHONPATH', ''))
    server = subprocess.Popen(os.path.join(build_path, 'src/server.py'), env=env)

    time.sleep(5)
    yield
    server.kill()


# pylint: disable=redefined-outer-name,unused-argument
def test_components(components, chrome_driver):
    """Test that no components cause an error."""
    chrome_driver.get('http://localhost:9991')
    chrome_driver.implicitly_wait(5)

    logs = chrome_driver.get_log('browser')
    for log in logs:
        if log['level'] == 'SEVERE':
            raise Exception(log['message'])



"""Test markdown and text widgets."""
# pylint: disable=redefined-outer-name,unused-argument,invalid-name

import os
from os import environ as env
import subprocess
import time

import pytest
from bowtie import App
from bowtie.html import Markdown
from bowtie.control import Textbox
from bowtie.tests.utils import reset_uuid


reset_uuid()

mark = Markdown('''
# top
## middle

[link]('hello.html')
''')
side = Markdown('''
# sideheader
''')
text = Textbox(area=True)


def write(txt):
    """Update markdown text."""
    mark.do_text(txt)


@pytest.fixture
def markdown(build_path, monkeypatch):
    """Create markdown and text widgets."""
    monkeypatch.setattr(App, '_sourcefile', lambda self: 'bowtie.tests.test_editor')

    app = App()
    app.add(mark)
    app.add_sidebar(side)
    app.add_sidebar(text)
    app.subscribe(write, text.on_change)
    # pylint: disable=protected-access
    app._build()

    env['PYTHONPATH'] = '{}:{}'.format(os.getcwd(), os.environ.get('PYTHONPATH', ''))
    server = subprocess.Popen(os.path.join(build_path, 'src/server.py'), env=env)

    time.sleep(5)
    yield
    server.kill()


def test_markdown(markdown, chrome_driver):
    """Test markdown and text widgets."""
    chrome_driver.get('http://localhost:9991')
    chrome_driver.implicitly_wait(5)

    assert chrome_driver.title == 'Bowtie App'

    txtctrl = chrome_driver.find_element_by_class_name('ant-input')
    output = chrome_driver.find_element_by_xpath(
        "//div[@style='grid-area: 1 / 2 / 2 / 3; position: relative;']"
    )

    assert 'top' in output.text
    assert 'middle' in output.text
    assert 'link' in output.text

    txtctrl.send_keys('apple')
    time.sleep(1)

    assert 'apple' in output.text

    txtctrl.send_keys('banana')
    time.sleep(1)

    assert 'apple' in output.text
    assert 'banana' in output.text



#!/usr/bin/env python
"""Example Bowtie App."""

from bowtie.control import Nouislider
from bowtie.visual import Plotly
from bowtie import Pager, cache, command

import numpy as np
from numpy import random as rng
import plotlywrapper as pw


pager = Pager()
sigma = Nouislider(start=0., minimum=0.1, maximum=50.)
mainplot = Plotly()


def initialize():
    cache.save('data', [0.] * 100)


def upgraph():
    data = cache.load('data')
    value = float(sigma.get())
    data.pop(0)
    data.append(value * rng.randn() + data[-1])
    mainplot.do_all(pw.line(data).to_json())
    cache.save('data', data)


def walk():
    pager.notify()


@command
def main():
    from bowtie import App
    app = App(debug=True)
    app.add_sidebar(sigma)
    app.add(mainplot)
    app.schedule(0.1, walk)
    app.respond(pager, upgraph)
    return app



"""Jupyter Integration."""

import ast
import socket
import os
from os.path import join as pjoin
from urllib.parse import urljoin
import json
import re
import sys
import types
import time
from subprocess import Popen, PIPE, STDOUT
from threading import Thread

from IPython import get_ipython
from IPython.display import display, HTML, clear_output
from IPython.core.error import UsageError
from IPython.core.interactiveshell import InteractiveShell
from IPython.core.magic import Magics, magics_class, line_magic
from nbformat import read
import ipykernel
import requests

from bowtie._app import _DIRECTORY, App


def get_notebook_name() -> str:
    """Return the full path of the jupyter notebook.

    References
    ----------
    https://github.com/jupyter/notebook/issues/1000#issuecomment-359875246

    """
    # this redefines a builtin >:( so putting it here to satisfy my linter
    from notebook.notebookapp import list_running_servers
    kernel_id = re.search(  # type: ignore
        'kernel-(.*).json',
        ipykernel.connect.get_connection_file()
    ).group(1)
    servers = list_running_servers()
    for server in servers:
        response = requests.get(urljoin(server['url'], 'api/sessions'),
                                params={'token': server.get('token', '')})
        for session in json.loads(response.text):
            if session['kernel']['id'] == kernel_id:
                relative_path = session['notebook']['path']
                return pjoin(server['notebook_dir'], relative_path)
    raise Exception('Noteboook not found.')


def load_notebook(fullname: str):
    """Import a notebook as a module."""
    shell = InteractiveShell.instance()
    path = fullname

    # load the notebook object
    with open(path, 'r', encoding='utf-8') as f:
        notebook = read(f, 4)

    # create the module and add it to sys.modules
    mod = types.ModuleType(fullname)
    mod.__file__ = path
    # mod.__loader__ = self
    mod.__dict__['get_ipython'] = get_ipython
    sys.modules[fullname] = mod

    # extra work to ensure that magics that would affect the user_ns
    # actually affect the notebook module's ns
    save_user_ns = shell.user_ns
    shell.user_ns = mod.__dict__

    try:
        for cell in notebook.cells:
            if cell.cell_type == 'code':
                try:
                    # only run valid python code
                    ast.parse(cell.source)
                except SyntaxError:
                    continue
                try:
                    # pylint: disable=exec-used
                    exec(cell.source, mod.__dict__)
                except NameError:
                    print(cell.source)
                    raise
    finally:
        shell.user_ns = save_user_ns
    return mod


@magics_class
class BowtieMagic(Magics):
    """Bowtie magic commands."""

    server = None

    @line_magic
    def bowtie_stop(self, line=''):  # pylint: disable=unused-argument
        """Terminate Bowtie app."""
        if self.server is None:
            print('No app has been run.')
        else:
            print('Terminating Bowtie app.')
            self.server.terminate()
            if self.server.poll() is None:
                time.sleep(1)
                if self.server.poll() is None:
                    print('Failed to stop Bowtie app.', file=sys.stderr)
                    return
            print('Successfully stopped Bowtie app.')
            self.server = None

    @line_magic
    def bowtie(self, line=''):
        """Build and serve a Bowtie app."""
        opts, appvar = self.parse_options(line, 'w:h:b:p:')
        width = opts.get('w', 1500)
        height = opts.get('h', 1000)
        border = opts.get('b', 2)
        port = opts.get('p', 9991)
        host = '0.0.0.0'

        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        result = sock.connect_ex((host, port))
        if result == 0:
            raise Exception('Port {} is unavailable on host {}, aborting.'.format(port, host))

        global_ns = self.shell.user_global_ns
        local_ns = self.shell.user_ns
        try:
            # pylint: disable=eval-used
            app = eval(appvar, global_ns, local_ns)
        except NameError:
            raise UsageError('Could not find App {}'.format(appvar))

        if not isinstance(app, App):
            raise UsageError('App is of type {} needs to be type <bowtie.App>'.format(type(app)))

        # pylint: disable=protected-access
        app._build(notebook=get_notebook_name())

        filepath = './{}/src/server.py'.format(_DIRECTORY)
        if os.path.isfile(filepath):
            self.server = Popen(['python', '-u', filepath], stdout=PIPE, stderr=STDOUT)
        else:
            raise FileNotFoundError('Cannot find "{}". Did you build the app?'.format(filepath))

        def flush_stdout(cmd):
            """Flush stdout from command continuously."""
            while True:
                line = cmd.stdout.readline()
                if line == b'' and cmd.poll() is not None:
                    return cmd.poll()
                print(line.decode('utf-8'), end='')
            raise Exception()

        thread = Thread(target=flush_stdout, args=(self.server,))
        thread.daemon = True
        thread.start()

        while self.server.poll() is None:
            try:
                if requests.get('http://localhost:9991').ok:
                    break
            except requests.exceptions.RequestException:
                continue
            time.sleep(1)
        else:
            print(self.server.stdout.read().decode('utf-8'), end='')

        clear_output()
        display(HTML(
            '<iframe src=http://localhost:9991 width={} height={} '
            'frameBorder={}></iframe>'.format(width, height, border)
        ))



"""Test markdown and text widgets."""

import time

import pytest
from bowtie import App, cache
from bowtie.control import Button
from bowtie.tests.utils import reset_uuid, server_check


reset_uuid()


# pylint: disable=invalid-name
button = Button()


def click():
    """Save and load cache."""
    cache['a'] = 3
    assert cache[b'a'] == 3
    assert cache[u'a'] == 3
    cache[b'b'] = True
    assert cache[b'b']
    assert cache[u'b']


def test_keys():
    """Test invalid keys."""
    with pytest.raises(KeyError):
        cache[0] = 0

    with pytest.raises(KeyError):
        cache[True] = 0

    with pytest.raises(KeyError):
        cache[3, 4] = 0


@pytest.fixture
def dummy(build_path, monkeypatch):
    """Create basic app."""
    monkeypatch.setattr(App, '_sourcefile', lambda self: 'bowtie.tests.test_cache')

    app = App()
    app.add(button)
    app.subscribe(click, button.on_click)
    # pylint: disable=protected-access
    app._build()

    with server_check(build_path) as server:
        yield server


# pylint: disable=redefined-outer-name,unused-argument
def test_cache(dummy, chrome_driver):
    """Test cache works."""
    chrome_driver.get('http://localhost:9991')
    chrome_driver.implicitly_wait(5)

    assert chrome_driver.title == 'Bowtie App'

    btx = chrome_driver.find_element_by_class_name('ant-btn')
    btx.click()
    time.sleep(2)



#!/usr/bin/env python
"""Command testing."""

from bowtie._command import numargs


def test_numargs():
    """Numargs testing."""

    # pylint: disable=missing-docstring
    def onearg(x):
        return x

    def zeroarg():
        pass

    def varargs(*y):
        return y

    def onevarargs(x, *y):
        return x, y

    assert numargs(onearg) == 1
    assert numargs(onevarargs) == 2
    assert numargs(zeroarg) == 0
    assert numargs(varargs) == 1



"""Utility functions for testing only."""

import os
from os.path import join as pjoin
from os import environ as env
from contextlib import contextmanager
from subprocess import Popen, PIPE
import time

from bowtie import View
from bowtie._component import Component


def reset_uuid():
    """Reset the uuid counter for components."""
    # pylint: disable=protected-access
    Component._NEXT_UUID = 0
    View._NEXT_UUID = 0


@contextmanager
def server_check(build_path):
    """Context manager for testing Bowtie apps and verifying no errors happened."""
    env['PYTHONPATH'] = '{}:{}'.format(os.getcwd(), os.environ.get('PYTHONPATH', ''))
    server = Popen(['python', '-u', pjoin(build_path, 'src/server.py')], env=env, stderr=PIPE)
    time.sleep(5)
    yield server
    server.terminate()
    _, stderr = server.communicate()
    assert b'Error' not in stderr
    assert b'Traceback' not in stderr



"""Plotly testing."""
# pylint: disable=redefined-outer-name,unused-argument,invalid-name

import os
from os import environ as env
import subprocess
import time

import pytest
from plotly import graph_objs as go

from bowtie import App
from bowtie.control import Nouislider, Button
from bowtie.visual import Plotly
from bowtie.tests.utils import reset_uuid


reset_uuid()

viz = Plotly()
ctrl = Nouislider()
ctrl_range = Nouislider(start=(20, 200))
ctrl2 = Button()


def callback(*args):
    """dummy function"""
    chart = go.Figure()
    chart.add_trace(go.Scatter(x=[1, 2, 3, 4], y=[4, 1, 3, 7]))
    chart.layout.title = 'hello world'
    viz.do_all(chart.to_dict())


@pytest.fixture
def plotly(build_path, monkeypatch):
    """Create plotly app."""
    monkeypatch.setattr(App, '_sourcefile', lambda self: 'bowtie.tests.test_plotly')

    app = App()
    app.add(viz)
    app.add_sidebar(ctrl)
    app.add_sidebar(ctrl_range)
    app.add_sidebar(ctrl2)
    app.subscribe(callback, ctrl.on_change)
    app.subscribe(callback, ctrl2.on_click)
    # pylint: disable=protected-access
    app._build()

    env['PYTHONPATH'] = '{}:{}'.format(os.getcwd(), os.environ.get('PYTHONPATH', ''))
    server = subprocess.Popen(os.path.join(build_path, 'src/server.py'), env=env)

    time.sleep(5)
    yield
    server.kill()


def test_plotly(plotly, chrome_driver):
    """Test plotly component."""
    chrome_driver.get('http://localhost:9991')
    chrome_driver.implicitly_wait(5)

    assert chrome_driver.title == 'Bowtie App'

    button = chrome_driver.find_element_by_class_name('ant-btn')
    points = chrome_driver.find_elements_by_class_name('point')
    assert not points

    button.click()

    points = chrome_driver.find_elements_by_class_name('point')

    logs = chrome_driver.get_log('browser')
    for log in logs:
        if log['level'] == 'SEVERE':
            raise Exception(log['message'])

    assert len(points) == 4



"""Defines the App class."""

from typing import (  # pylint: disable=unused-import
    Any, Callable, Generator, List, Optional, Set, Tuple, Union, Dict, Sequence
)
import os
import json
import itertools
import inspect
import shutil
import stat
from collections import namedtuple, defaultdict
from subprocess import Popen, PIPE, STDOUT
from pathlib import Path
import warnings

from jinja2 import Environment, FileSystemLoader

from bowtie._component import Event, Component, COMPONENT_REGISTRY
from bowtie.pager import Pager
from bowtie.exceptions import (
    GridIndexError, NoSidebarError,
    NotStatefulEvent, NoUnusedCellsError,
    SpanOverlapError, SizeError, WebpackError, YarnError
)


Route = namedtuple('Route', ['view', 'path', 'exact'])
_Import = namedtuple('_Import', ['module', 'component'])
_Schedule = namedtuple('_Schedule', ['seconds', 'function'])

_DIRECTORY = Path('build')
_WEBPACK = './node_modules/.bin/webpack'


def raise_not_number(x: int) -> None:
    """Raise ``SizeError`` if ``x`` is not a number``."""
    try:
        float(x)
    except ValueError:
        raise SizeError('Must pass a number, received {}'.format(x))


class Span:
    """Define the location of a widget."""

    def __init__(self, row_start: int, column_start: int, row_end: Optional[int] = None,
                 column_end: Optional[int] = None) -> None:
        """Create a span for a widget.

        Indexing starts at 0. Start is inclusive and end is exclusive

        CSS Grid indexing starts at 1 and is [inclusive, exclusive)

        Note: `_start` and `_end` follow css grid naming convention.

        Parameters
        ----------
        row_start : int
        column_start : int
        row_end : int, optional
        column_end : int, optional

        """
        self.row_start = row_start
        self.column_start = column_start
        # add 1 to then ends because they start counting from 1
        if row_end is None:
            self.row_end = self.row_start + 1
        else:
            self.row_end = row_end
        if column_end is None:
            self.column_end = self.column_start + 1
        else:
            self.column_end = column_end

    @property
    def _key(self) -> Tuple[int, int, int, int]:
        return self.row_start, self.column_start, self.row_end, self.column_end

    def __hash__(self) -> int:
        """Hash for dict."""
        return hash(self._key)

    def __eq__(self, other) -> bool:
        """Compare eq for dict."""
        # pylint: disable=protected-access
        return isinstance(other, type(self)) and self._key == other._key

    def __repr__(self) -> str:
        """Show the starting and ending points.

        This is used as a key in javascript.
        """
        return '{},{},{},{}'.format(
            self.row_start + 1,
            self.column_start + 1,
            self.row_end + 1,
            self.column_end + 1
        )

    def overlap(self, other: 'Span'):
        """Detect if two spans overlap."""
        return not (
            # if one rectangle is left of other
            other.column_end <= self.column_start
            or self.column_end <= other.column_start
            # if one rectangle is above other
            or other.row_end <= self.row_start
            or self.row_end <= other.row_start
        )

    @property
    def cells(self) -> Generator[Tuple[int, int], None, None]:
        """Generate cells in span."""
        yield from itertools.product(
            range(self.row_start, self.row_end),
            range(self.column_start, self.column_end)
        )


class Size:
    """Size of rows and columns in grid.

    This is accessed through ``.rows`` and ``.columns`` from App and View instances.

    This uses CSS's minmax function.

    The minmax() CSS function defines a size range greater than or equal
    to min and less than or equal to max. If max < min, then max is ignored
    and minmax(min,max) is treated as min. As a maximum, a <flex> value
    sets the flex factor of a grid track; it is invalid as a minimum.

    Examples
    --------
    Laying out an app with the first row using 1/3 of the space
    and the second row using 2/3 of the space.

    >>> app = App(rows=2, columns=3)
    >>> app.rows[0].fraction(1)
    1fr
    >>> app.rows[1].fraction(2)
    2fr

    """

    def __init__(self) -> None:
        """Create a default row or column size with fraction = 1."""
        self.minimum: str = ''
        self.maximum: str = ''
        self.fraction(1)

    def auto(self) -> 'Size':
        """Set the size to auto or content based."""
        self.maximum = 'auto'
        return self

    def min_auto(self) -> 'Size':
        """Set the minimum size to auto or content based."""
        self.minimum = 'auto'
        return self

    def pixels(self, value) -> 'Size':
        """Set the size in pixels."""
        raise_not_number(value)
        self.maximum = '{}px'.format(value)
        return self

    def min_pixels(self, value) -> 'Size':
        """Set the minimum size in pixels."""
        raise_not_number(value)
        self.minimum = '{}px'.format(value)
        return self

    def ems(self, value) -> 'Size':
        """Set the size in ems."""
        raise_not_number(value)
        self.maximum = '{}em'.format(value)
        return self

    def min_ems(self, value) -> 'Size':
        """Set the minimum size in ems."""
        raise_not_number(value)
        self.minimum = '{}em'.format(value)
        return self

    def fraction(self, value: int) -> 'Size':
        """Set the fraction of free space to use as an integer."""
        raise_not_number(value)
        self.maximum = '{}fr'.format(int(value))
        return self

    def percent(self, value) -> 'Size':
        """Set the percentage of free space to use."""
        raise_not_number(value)
        self.maximum = '{}%'.format(value)
        return self

    def min_percent(self, value) -> 'Size':
        """Set the minimum percentage of free space to use."""
        raise_not_number(value)
        self.minimum = '{}%'.format(value)
        return self

    def __repr__(self) -> str:
        """Represent the size to be inserted into a JSX template."""
        if self.minimum:
            return 'minmax({}, {})'.format(self.minimum, self.maximum)
        return self.maximum


class Gap:
    """Margin between rows or columns of the grid.

    This is accessed through ``.row_gap`` and ``.column_gap`` from App and View instances.

    Examples
    --------
    Create a gap of 5 pixels between all rows.

    >>> app = App()
    >>> app.row_gap.pixels(5)
    5px

    """

    def __init__(self) -> None:
        """Create a default margin of zero."""
        self.gap: str = ''
        self.pixels(0)

    def pixels(self, value: int) -> 'Gap':
        """Set the margin in pixels."""
        raise_not_number(value)
        self.gap = '{}px'.format(value)
        return self

    def ems(self, value: int) -> 'Gap':
        """Set the margin in ems."""
        raise_not_number(value)
        self.gap = '{}em'.format(value)
        return self

    def percent(self, value) -> 'Gap':
        """Set the margin as a percentage."""
        raise_not_number(value)
        self.gap = '{}%'.format(value)
        return self

    def __repr__(self) -> str:
        """Represent the margin to be inserted into a JSX template."""
        return self.gap


def _check_index(value: int, length: int, bound: bool) -> int:
    if not isinstance(value, int):
        raise GridIndexError('Indices must be integers, found {}.'.format(value))
    if value < 0:
        value = value + length
    if value < 0 + bound or value >= length + bound:
        raise GridIndexError('Index out of range.')
    return value


def _slice_to_start_end(slc: slice, length: int) -> Tuple[int, int]:
    if slc.step is not None and slc.step != 1:
        raise GridIndexError(
            'slice step is not supported must be None or 1, was {}'.format(slc.step)
        )

    start = 0
    if slc.start is not None:
        start = slc.start

    end = length
    if slc.stop is not None:
        end = slc.stop
    return start, end


class Components:
    """List like class for storing components to override iadd.

    The purpose of this class is to override the `iadd` function.
    I want to be able to support all the following
    >>> from bowtie import App
    >>> from bowtie.control import Button
    >>> app = App()
    >>> button = Button()
    >>> app[0, 0] = button
    >>> app[0, 0] = button, button
    >>> app[0, 0] += button
    >>> app[0, 0] += button, button
    """

    TYPE_MSG: str = 'Must add a component or sequence of components, found {}.'

    def __init__(self,
                 component: Optional[Union[Component, Sequence[Component]]] = None
                 ) -> None:
        """Create a components list."""
        self.data: List[Component]
        if component is None:
            self.data = []
        elif isinstance(component, Component):
            self.data = [component]
        else:
            self.data = list(component)

    def __len__(self):
        """Count components."""
        return self.data.__len__()

    def append(self, component: Component):
        """Append component to the list."""
        return self.data.append(component)

    def __iter__(self):
        """Iterate over components."""
        return self.data.__iter__()

    def __getitem__(self, key):
        """Get item as a list."""
        return self.data.__getitem__(key)

    def _add(self, method, other: Union[Component, Sequence[Component]]) -> 'Components':
        if isinstance(other, Component):
            return method([other])
        if isinstance(other, Sequence):
            other = list(other)
            if not all(True for x in other if isinstance(x, Component)):
                raise TypeError(self.TYPE_MSG.format(other))
            return method(other)
        raise TypeError(self.TYPE_MSG.format(other))

    def __iadd__(self, other: Union[Component, Sequence[Component]]):
        """Append items to list when adding."""
        return self._add(self.data.__iadd__, other)

    def __add__(self, other: Union[Component, Sequence[Component]]):
        """Append items to list when adding."""
        return self._add(self.data.__add__, other)


class View:
    """Grid of components."""

    _NEXT_UUID = 0

    @classmethod
    def _next_uuid(cls) -> int:
        cls._NEXT_UUID += 1
        return cls._NEXT_UUID

    def __init__(self, rows: int = 1, columns: int = 1, sidebar: bool = True,
                 background_color: str = 'White') -> None:
        """Create a new grid.

        Parameters
        ----------
        rows : int, optional
            Number of rows in the grid.
        columns : int, optional
            Number of columns in the grid.
        sidebar : bool, optional
            Enable a sidebar for control components.
        background_color : str, optional
            Background color of the control pane.

        """
        self._uuid = View._next_uuid()
        self.column_gap = Gap()
        self.row_gap = Gap()
        self.border = Gap().pixels(7)
        self.rows = [Size() for _ in range(rows)]
        self.columns = [Size() for _ in range(columns)]
        self.sidebar = sidebar
        self.background_color = background_color
        self._controllers: List[Component] = []
        self._spans: Dict[Span, Components] = {}

    def _all_components(self) -> Generator[Component, None, None]:
        yield from self._controllers
        yield from itertools.chain.from_iterable(self._spans.values())

    @property
    def _packages(self) -> Set[str]:
        # pylint: disable=protected-access
        packages = set(x._PACKAGE for x in self._all_components())
        packages.discard(None)
        return packages

    @property
    def _templates(self) -> Set[str]:
        # pylint: disable=protected-access
        return set(x._TEMPLATE for x in self._all_components())

    @property
    def _imports(self) -> Set[_Import]:
        # pylint: disable=protected-access
        return set(_Import(component=x._COMPONENT,
                           module=x._TEMPLATE[:x._TEMPLATE.find('.')])
                   for x in self._all_components())

    @property
    def _components(self) -> Set[Component]:
        return set(self._all_components())

    def _key_to_span(self, key: Any) -> Span:
        # TODO spaghetti code cleanup needed!
        if isinstance(key, Span):
            return key
        if isinstance(key, tuple):
            if len(key) == 1:
                return self._key_to_span(key[0])
            try:
                row_key, column_key = key
            except ValueError:
                raise GridIndexError('Index must be 1 or 2 values, found {}'.format(key))
            if isinstance(row_key, int):
                row_start = _check_index(row_key, len(self.rows), False)
                row_end = row_start + 1
            elif isinstance(row_key, slice):
                row_start, row_end = _slice_to_start_end(row_key, len(self.rows))
                row_start = _check_index(row_start, len(self.rows), False)
                row_end = _check_index(row_end, len(self.rows), True)
            else:
                raise GridIndexError(
                    'Cannot index with {}, pass in a int or a slice.'.format(row_key)
                )

            if isinstance(column_key, int):
                column_start = _check_index(column_key, len(self.columns), False)
                column_end = column_start + 1
            elif isinstance(column_key, slice):
                column_start, column_end = _slice_to_start_end(column_key, len(self.columns))
                column_start = _check_index(column_start, len(self.columns), False)
                column_end = _check_index(column_end, len(self.columns), True)
            else:
                raise GridIndexError(
                    'Cannot index with {}, pass in a int or a slice.'.format(column_key)
                )
            rows_cols = row_start, column_start, row_end, column_end

        elif isinstance(key, slice):
            start, end = _slice_to_start_end(key, len(self.rows))
            start = _check_index(start, len(self.rows), False)
            end = _check_index(end, len(self.rows), True)
            rows_cols = start, 0, end, len(self.columns)
        elif isinstance(key, int):
            row_start = _check_index(key, len(self.rows), False)
            rows_cols = row_start, 0, row_start + 1, len(self.columns)
        else:
            raise GridIndexError('Invalid index {}'.format(key))
        return Span(*rows_cols)

    def __getitem__(self, key: Any) -> Components:
        """Get item from the view."""
        span = self._key_to_span(key)
        if span not in self._spans:
            raise KeyError(f'Key {key} has not been used')
        return self._spans[span]

    def __setitem__(self, key: Any,
                    component: Union[Component, Sequence[Component]]) -> None:
        """Add widget to the view."""
        span = self._key_to_span(key)
        for used_span in self._spans:
            if span != used_span and span.overlap(used_span):
                raise SpanOverlapError(f'Spans {span} and {used_span} overlap. '
                                       'This is not permitted. '
                                       'If you want to do this please open an issue '
                                       'and explain your use case. '
                                       'https://github.com/jwkvam/bowtie/issues')
        self._spans[span] = Components(component)

    def add(self, component: Union[Component, Sequence[Component]]) -> None:
        """Add a widget to the grid in the next available cell.

        Searches over columns then rows for available cells.

        Parameters
        ----------
        components : bowtie._Component
            A Bowtie widget instance.

        """
        try:
            self[Span(*self._available_cell())] = component
        except NoUnusedCellsError:
            span = list(self._spans.keys())[-1]
            self._spans[span] += component

    def _available_cell(self) -> Tuple[int, int]:
        """Find next available cell first by row then column.

        First, construct a set containing all cells.
        Then iterate over the spans and remove occupied cells.
        """
        cells = set(itertools.product(range(len(self.rows)), range(len(self.columns))))
        for span in self._spans:
            for cell in span.cells:
                cells.remove(cell)

        if not cells:
            raise NoUnusedCellsError('No available cells')
        return min(cells)

    def add_sidebar(self, component: Component) -> None:
        """Add a widget to the sidebar.

        Parameters
        ----------
        component : bowtie._Component
            Add this component to the sidebar, it will be appended to the end.

        """
        if not self.sidebar:
            raise NoSidebarError('Set `sidebar=True` if you want to use the sidebar.')

        if not isinstance(component, Component):
            raise ValueError('component must be Component type, found {}'.format(component))
        # self._track_widget(widget)
        self._controllers.append(component)  # pylint: disable=protected-access

    @property
    def _columns_sidebar(self):
        columns = []
        if self.sidebar:
            columns.append(Size().ems(18))
        columns += self.columns
        return columns


class App:
    """Core class to layout, connect, build a Bowtie app."""

    def __init__(self, rows: int = 1, columns: int = 1, sidebar: bool = True,
                 title: str = 'Bowtie App', basic_auth: bool = False,
                 username: str = 'username', password: str = 'password',
                 theme: Optional[str] = None, background_color: str = 'White',
                 host: str = '0.0.0.0', port: int = 9991, socketio: str = '',
                 debug: bool = False) -> None:
        """Create a Bowtie App.

        Parameters
        ----------
        row : int, optional
            Number of rows in the grid.
        columns : int, optional
            Number of columns in the grid.
        sidebar : bool, optional
            Enable a sidebar for control components.
        title : str, optional
            Title of the HTML.
        basic_auth : bool, optional
            Enable basic authentication.
        username : str, optional
            Username for basic authentication.
        password : str, optional
            Password for basic authentication.
        theme : str, optional
            Color for Ant Design components.
        background_color : str, optional
            Background color of the control pane.
        host : str, optional
            Host IP address.
        port : int, optional
            Host port number.
        socketio : string, optional
            Socket.io path prefix, only change this for advanced deployments.
        debug : bool, optional
            Enable debugging in Flask. Disable in production!

        """
        self._basic_auth = basic_auth
        self._debug = debug
        self._host = host
        self._init: Optional[str] = None
        self._password = password
        self._port = port
        self._socketio = socketio
        self._schedules: List[_Schedule] = []
        self._subscriptions: Dict[Event, List[Tuple[List[Event], str]]] = defaultdict(list)
        self._pages: Dict[Pager, str] = {}
        self._title = title
        self._username = username
        self._uploads: Dict[int, str] = {}
        self.theme = theme
        self._root = View(rows=rows, columns=columns, sidebar=sidebar,
                          background_color=background_color)
        self._routes = [Route(view=self._root, path='/', exact=True)]
        self._package_dir = Path(os.path.dirname(__file__))
        self._jinjaenv = Environment(
            loader=FileSystemLoader(str(self._package_dir / 'templates')),
            trim_blocks=True,
            lstrip_blocks=True
        )

    def __getattr__(self, name: str) -> Union[Gap, List[Size]]:
        """Export attributes from root view."""
        if name == 'columns':
            return self._root.columns
        if name == 'rows':
            return self._root.rows
        if name == 'column_gap':
            return self._root.column_gap
        if name == 'row_gap':
            return self._root.row_gap
        if name == 'border':
            return self._root.border
        raise AttributeError(name)

    def __getitem__(self, key: Any):
        """Get item from root view."""
        return self._root.__getitem__(key)

    def __setitem__(self, key: Any,
                    value: Union[Component, Sequence[Component]]) -> None:
        """Add widget to the root view."""
        self._root.__setitem__(key, value)

    def add(self, component: Component) -> None:
        """Add a widget to the grid in the next available cell.

        Searches over columns then rows for available cells.

        Parameters
        ----------
        component : bowtie._Component
            A Bowtie component instance.

        """
        self._root.add(component)

    def add_sidebar(self, widget: Component) -> None:
        """Add a widget to the sidebar.

        Parameters
        ----------
        widget : bowtie._Component
            Add this widget to the sidebar, it will be appended to the end.

        """
        self._root.add_sidebar(widget)

    def add_route(self, view: View, path: str, exact: bool = True) -> None:
        """Add a view to the app.

        Parameters
        ----------
        view : View
        path : str
        exact : bool, optional

        """
        if path[0] != '/':
            path = '/' + path
        for route in self._routes:
            assert path != route.path, 'Cannot use the same path twice'
        self._routes.append(Route(view=view, path=path, exact=exact))

    def respond(self, pager: Pager, func: Callable) -> None:
        """Call a function in response to a page.

        When the pager calls notify, the function will be called.

        Parameters
        ----------
        pager : Pager
            Pager that to signal when func is called.
        func : callable
            Function to be called.

        Examples
        --------
        Using the pager to run a callback function.

        >>> from bowtie.pager import Pager
        >>> app = App()
        >>> pager = Pager()
        >>> def callback():
        ...     pass
        >>> def scheduledtask():
        ...     pager.notify()
        >>> app.respond(pager, callback)

        """
        self._pages[pager] = func.__name__

    def subscribe(self, func: Callable, event: Event, *events: Event) -> None:
        """Call a function in response to an event.

        If more than one event is given, `func` will be given
        as many arguments as there are events.

        Parameters
        ----------
        func : callable
            Function to be called.
        event : event
            A Bowtie event.
        *events : Each is an event, optional
            Additional events.

        Examples
        --------
        Subscribing a function to multiple events.

        >>> from bowtie.control import Dropdown, Slider
        >>> app = App()
        >>> dd = Dropdown()
        >>> slide = Slider()
        >>> def callback(dd_item, slide_value):
        ...     pass
        >>> app.subscribe(callback, dd.on_change, slide.on_change)

        """
        if not callable(func):
            raise TypeError(
                'The first argument to subscribe must be callable, found {}'.format(type(func))
            )
        all_events = [event, *events]
        if len(all_events) != len(set(all_events)):
            raise ValueError('Subscribed to the same event multiple times. '
                             'All events must be unique.')

        if len(all_events) > 1:
            # check if we are using any non stateful events
            for evt in all_events:
                if evt.getter is None:
                    msg = '{}.on_{} is not a stateful event. It must be used alone.'
                    raise NotStatefulEvent(msg.format(evt.uuid, evt.name))

        if event.name == 'upload':
            if event.uuid in self._uploads:
                warnings.warn(
                    ('Overwriting function "{func1}" with function '
                     '"{func2}" for upload object "{obj}".').format(
                         func1=self._uploads[event.uuid],
                         func2=func.__name__,
                         obj=COMPONENT_REGISTRY[event.uuid]
                     ), Warning)
            self._uploads[event.uuid] = func.__name__

        for evt in all_events:
            self._subscriptions[evt].append((all_events, func.__name__))

    def listen(self, event: Event, *events: Event) -> Callable:
        """Call a function in response to an event.

        If more than one event is given, `func` will be given
        as many arguments as there are events.

        Parameters
        ----------
        event : event
            A Bowtie event.
        *events : Each is an event, optional
            Additional events.

        Examples
        --------
        Subscribing a function to multiple events.

        >>> from bowtie.control import Dropdown, Slider
        >>> app = App()
        >>> dd = Dropdown()
        >>> slide = Slider()
        >>> @app.listen(dd.on_change, slide.on_change)
        ... def callback(dd_item, slide_value):
        ...     pass
        >>> @app.listen(dd.on_change)
        ... @app.listen(slide.on_change)
        ... def callback2(value):
        ...     pass

        """
        def decorator(func):
            """Subscribe function to events."""
            self.subscribe(func, event, *events)
            return func
        return decorator

    def load(self, func: Callable) -> None:
        """Call a function on page load.

        Parameters
        ----------
        func : callable
            Function to be called.

        """
        self._init = func.__name__

    def schedule(self, seconds: float, func: Callable) -> None:
        """Call a function periodically.

        Parameters
        ----------
        seconds : float
            Minimum interval of function calls.
        func : callable
            Function to be called.

        """
        self._schedules.append(_Schedule(seconds, func.__name__))

    def _sourcefile(self) -> str:  # pylint: disable=no-self-use
        # [-1] grabs the top of the stack
        return os.path.basename(inspect.stack()[-1].filename)[:-3]

    def _write_templates(self, notebook: Optional[str] = None) -> Set[str]:
        server = self._jinjaenv.get_template('server.py.j2')
        indexhtml = self._jinjaenv.get_template('index.html.j2')
        indexjsx = self._jinjaenv.get_template('index.jsx.j2')
        componentsjs = self._jinjaenv.get_template('components.js.j2')
        webpack = self._jinjaenv.get_template('webpack.common.js.j2')

        src, app, templates = create_directories()

        webpack_path = _DIRECTORY / webpack.name[:-3]  # type: ignore
        with webpack_path.open('w') as f:
            f.write(
                webpack.render(color=self.theme)
            )

        server_path = src / server.name[:-3]  # type: ignore
        with server_path.open('w') as f:
            f.write(
                server.render(
                    basic_auth=self._basic_auth,
                    username=self._username,
                    password=self._password,
                    notebook=notebook,
                    source_module=self._sourcefile() if not notebook else None,
                    subscriptions=self._subscriptions,
                    uploads=self._uploads,
                    schedules=self._schedules,
                    initial=self._init,
                    routes=self._routes,
                    pages=self._pages,
                    host="'{}'".format(self._host),
                    port=self._port,
                    debug=self._debug
                )
            )

        perms = os.stat(server_path)
        os.chmod(server_path, perms.st_mode | stat.S_IEXEC)

        # copy js modules that are always needed
        for name in ['progress.jsx', 'view.jsx', 'utils.js']:
            template_src = self._package_dir / 'src' / name
            shutil.copy(template_src, app)

        for route in self._routes:
            # pylint: disable=protected-access
            for template in route.view._templates:
                template_src = self._package_dir / 'src' / template
                shutil.copy(template_src, app)

        # Layout Design
        #
        # Dictionaries that are keyed by the components
        #
        # To layout this will need to look through all components that have a key of the route
        #
        #
        # 1. This way they can

        # use cases
        # 1. statically add items to controller in list
        # 2. remove item from controller
        # 3. add item back to controller
        #
        # issues:
        # widget reordering
        # order preserving operations

        components: Set[Component] = set()
        imports: Set[_Import] = set()
        packages: Set[str] = set()
        for route in self._routes:
            packages |= route.view._packages  # pylint: disable=protected-access
            imports |= route.view._imports  # pylint: disable=protected-access
            components |= route.view._components  # pylint: disable=protected-access

        with (app / componentsjs.name[:-3]).open('w') as f:  # type: ignore
            f.write(
                componentsjs.render(
                    imports=imports,
                    socketio=self._socketio,
                    components=components,
                )
            )

        with (templates / indexhtml.name[:-3]).open('w') as f:  # type: ignore
            f.write(
                indexhtml.render(
                    title=self._title,
                )
            )

        with (app / indexjsx.name[:-3]).open('w') as f:  # type: ignore
            f.write(
                indexjsx.render(
                    maxviewid=View._NEXT_UUID,  # pylint: disable=protected-access
                    socketio=self._socketio,
                    pages=self._pages,
                    routes=self._routes,
                )
            )
        return packages

    def _build(self, notebook: Optional[str] = None) -> None:
        """Compile the Bowtie application."""
        packages = self._write_templates(notebook=notebook)

        if not os.path.isfile(os.path.join(_DIRECTORY, 'package.json')):
            packagejson = os.path.join(self._package_dir, 'src/package.json')
            shutil.copy(packagejson, _DIRECTORY)

        if not os.path.isfile(os.path.join(_DIRECTORY, 'webpack.prod.json')):
            webpackprod = os.path.join(self._package_dir, 'src/webpack.prod.js')
            shutil.copy(webpackprod, _DIRECTORY)

        if not os.path.isfile(os.path.join(_DIRECTORY, 'webpack.dev.json')):
            webpackdev = os.path.join(self._package_dir, 'src/webpack.dev.js')
            shutil.copy(webpackdev, _DIRECTORY)

        if run(['yarn', '--ignore-engines', 'install'], notebook=notebook) > 1:
            raise YarnError('Error installing node packages')

        if packages:
            installed = installed_packages()
            new_packages = [x for x in packages if x.split('@')[0] not in installed]

            if new_packages:
                retval = run(['yarn', '--ignore-engines', 'add'] + new_packages, notebook=notebook)
                if retval > 1:
                    raise YarnError('Error installing node packages')
                elif retval == 1:
                    print('Yarn error but trying to continue build')
        retval = run([_WEBPACK, '--config', 'webpack.dev.js'], notebook=notebook)
        if retval != 0:
            raise WebpackError('Error building with webpack')


def run(command: List[str], notebook: Optional[str] = None) -> int:
    """Run command from terminal and notebook and view output from subprocess."""
    if notebook is None:
        return Popen(command, cwd=_DIRECTORY).wait()
    cmd = Popen(command, cwd=_DIRECTORY, stdout=PIPE, stderr=STDOUT)
    while True:
        line = cmd.stdout.readline()
        if line == b'' and cmd.poll() is not None:
            return cmd.poll()
        print(line.decode('utf-8'), end='')
    raise Exception()


def installed_packages() -> Generator[str, None, None]:
    """Extract installed packages as list from `package.json`."""
    with (_DIRECTORY / 'package.json').open('r') as f:
        packages = json.load(f)
    yield from packages['dependencies'].keys()


def create_directories() -> Tuple[Path, Path, Path]:
    """Create all the necessary subdirectories for the build."""
    src = _DIRECTORY / 'src'
    templates = src / 'templates'
    app = src / 'app'
    os.makedirs(app, exist_ok=True)
    os.makedirs(templates, exist_ok=True)
    return src, app, templates



"""Bowtie exceptions."""


class YarnError(Exception):
    """Errors from ``Yarn``."""

    pass


class WebpackError(Exception):
    """Errors from ``Webpack``."""

    pass


class SizeError(Exception):
    """Size values must be a number."""

    pass


class GridIndexError(IndexError):
    """Invalid index into the grid layout."""

    pass


class NoUnusedCellsError(Exception):
    """All cells are used."""

    pass


class SpanOverlapError(Exception):
    """Spans may not overlap."""

    pass


class MissingRowOrColumn(Exception):
    """Missing a row or column."""

    pass


class NoSidebarError(Exception):
    """Cannot add to the sidebar when it doesn't exist."""

    pass


class NotStatefulEvent(Exception):
    """This event is not stateful and cannot be paired with other events."""

    pass


class SerializationError(TypeError):
    """Cannot serialize the data for command."""

    pass
